/* "commands.c" copyright 1994 thomas insel
 *              copyright 1995 sven oliver moll */

#include "config.h"
#include "common.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include "commands.h"
#include "hardware.h"
#include "shuffle.h"
#include "util.h"
#include "cdtool.h"

/************************************************************************/
/* Procedure:  do_play
 * Purpose:    to start playing
 * 
 * Inputs:     program name, CD file des., initial track, last track
 * Outputs:    to CD-ROM, to stdout
 * Returns:    0< on error; 0 (TRUE) on success
 * Notes:  
 *   1.
 */
/************************************************************************/
/* if trk0 is -1, skip back
 * if trk0 is -2, skip forward
 */
int do_play(cdtool_t *cdt, int cdfd, int trk0, int trk1, cdhw_t *hw)
{
    struct cdrom_tochdr tochdr;
    struct cdrom_ti ti;
    struct cdrom_subchnl subchnl;

    DEBUGMSG("do_play:  called, name=%s, cdfd=%d, trk0=%d, trk1=%d",
      cdt->progname, cdfd, trk0, trk1);

    subchnl.cdsc_format = CDROM_MSF;

    if (((trk0 == 0) && (trk1 == 0)) || (trk0 < 0)) {
        DEBUGMSG("do_play:  trk0=%d", trk0);

        if (ioctl(cdfd, CDROMSUBCHNL, &subchnl) == -1) {
            ERRORMSG("%s: ioctl cdromsubchnl: %s", cdt->progname,
              strerror(errno));
            return(-1);
        }
    }

    /* if starting play.... */
    if ((trk0 == 0) && (trk1 == 0)) {
        DEBUGMSG("do_play:  trk[0.1]==0, status=%s",
          (subchnl.cdsc_audiostatus == CDROM_AUDIO_PAUSED) ? "PAUSED":"PLAY");

        /* subchannel status returns PLAY, PAUSED, COMPLETED, or none */
        switch (subchnl.cdsc_audiostatus) {
            case CDROM_AUDIO_PAUSED:
                if (ioctl(cdfd, CDROMRESUME) == -1) {
                    ERRORMSG("%s: ioctl cdromresume: %s", cdt->progname,
                      strerror(errno));
                    return(-1);
                }
                return(0);

            case CDROM_AUDIO_PLAY:
                trk0=subchnl.cdsc_trk;
                break;

            case CDROM_AUDIO_COMPLETED:  /* WWH added */
                trk0=1;
                break;

#ifdef USE_THIS
/* TODO *  what is this? */
            default:
                ERRORMSG("%s: error subchnl=%d",
                        cdt->progname, subchnl.cdsc_audiostatus);
                return(-1);
#endif	    
        }
    }

    if (ioctl(cdfd, CDROMREADTOCHDR, &tochdr) == -1) {
        ERRORMSG("%s: ioctl cdromreadtochdr: %s", cdt->progname,
          strerror(errno));
        return(-1);
    }

    /* TODO
     *  move this calculation into it's own function
     */

    /* handle previous and next track */
    if (trk0 < 0)
    {
        if (trk1 == 0)
            trk1 = 1;

        switch (trk0) {
            case SKIPBACK:
                trk1 = subchnl.cdsc_trk - trk1;
                break;
            case SKIPFORWARD:
                trk1 = subchnl.cdsc_trk + trk1;
        }

        if ((trk1 < tochdr.cdth_trk0) || (trk1 > tochdr.cdth_trk1))
            trk1=1;
        ti.cdti_trk0 = trk1;
        ti.cdti_trk1 = tochdr.cdth_trk1;
    }
    else
    {
        if (trk0) 
            ti.cdti_trk0 = trk0;
        else
            ti.cdti_trk0 = tochdr.cdth_trk0;

        if (trk1)
            ti.cdti_trk1 = trk1;
        else
            ti.cdti_trk1 = tochdr.cdth_trk1;
    }

    ti.cdti_ind0 = 1;
    ti.cdti_ind1 = 99;

    /* now, adjust the desired playing range to compensate for DATA tracks */
    if (hw)
    {
        int first = first_track(hw);
        int last = last_track(hw);
        int newlast, rest, warned;

        if (ti.cdti_trk0 < first)
            ti.cdti_trk0 = first;
        if (ti.cdti_trk1 > last)
            ti.cdti_trk1 = last;

        for (; ti.cdti_trk0 <= ti.cdti_trk1; ti.cdti_trk0++)
            if (is_audio_track(hw, ti.cdti_trk0))
                break;

        if (ti.cdti_trk0 > ti.cdti_trk1) {
            ERRORMSG("%s: no audio tracks within requested range",
              cdt->progname);
            return(-1);
        }

        for (newlast = ti.cdti_trk0; newlast < ti.cdti_trk1; newlast++)
            if (!is_audio_track(hw, newlast + 1))
                break;

        for (rest=newlast+1, warned=0; rest <= ti.cdti_trk1; rest++)
            if (is_audio_track(hw, rest)) {
                if (!warned) {
                    ERRORMSG("%s: warning, only playing to first "
                      "data track in range", cdt->progname);
                    fprintf(stderr, "   audio tracks");
                    warned = 1;
                }
                fprintf(stderr, " %d", rest);
            }

        if (warned)
            ERRORMSG(" will be missed");

        ti.cdti_trk1 = newlast;
    }

    VERBOSEMSG("Playing from track %d to %d", ti.cdti_trk0, ti.cdti_trk1);

    /* check range */
    DEBUGMSG("do_play: ti.cdti_trk0=%d, ti.cdti_trk1=%d",
      ti.cdti_trk0, ti.cdti_trk1);

    /* make this a runtime configuration? */
#ifndef SCSI
    if ( ioctl(cdfd, CDROMPLAYTRKIND, &ti) == -1 ) 
    {
        /* patch by <Ingo.Wilken@Informatik.Uni-Oldenburg.DE> for Plextor */
        /* PX-32TSI drive */
        /* retry with all indices set to 1 */
        ti.cdti_ind1 = 1;
        if ( ioctl(cdfd, CDROMPLAYTRKIND, &ti) == -1 )
        {
            ERRORMSG("%s: ioctl cdromplaytrkind: %s", cdt->progname,
                    strerror(errno));
            return(-1);
        }
    }
#else
    /* SCSI - convert track index into MSF format */
    {
        cdhw_t *hw;
        struct cdrom_msf msf;
        int i;

        if ((hw = malloc(sizeof(cdhw_t))) == NULL)
            return(-1);

        if (read_hw(cdt, &hw, cdfd) != TRUE || hw == NULL) {
            ERRORMSG("%s: %s", cdt->progname,
              cd_status_text(hw->subchnl.cdsc_audiostatus));
            return(-1);
        }

        if (ioctl(cdfd, CDROMSTART) == -1 ) {
            free_hw_buf(hw);
            ERRORMSG("%s: ioctl cdromstart: %s", cdt->progname,
              strerror(errno));
            return(-1);
        }

        /* convert track offset to MSF format */
        i = ti.cdti_trk0;
        if (i > hw->tochdr.cdth_trk1) {
            free_hw_buf(hw);
            ERRORMSG("%s: invalid track %d", cdt->progname, i);
            return(-1);
        }

        msf.cdmsf_min0 = hw->tocentries[i-1].cdte_addr.msf.minute;
        msf.cdmsf_sec0 = hw->tocentries[i-1].cdte_addr.msf.second;
        msf.cdmsf_frame0 = hw->tocentries[i-1].cdte_addr.msf.frame;

        i = ti.cdti_trk1;
        if (i == 0)
            i = hw->tochdr.cdth_trk1+1;
        else
            i++;

        msf.cdmsf_min1 = hw->tocentries[i-1].cdte_addr.msf.minute;
        msf.cdmsf_sec1 = hw->tocentries[i-1].cdte_addr.msf.second;
        msf.cdmsf_frame1 = hw->tocentries[i-1].cdte_addr.msf.frame;

        free_hw_buf(hw);
        if (ioctl(cdfd, CDROMPLAYMSF, &msf) == -1) {
            ERRORMSG("%s: ioctl cdrommsf: %s", cdt->progname, strerror(errno));
            return(-1);
        }
    }
#endif
    return(TRUE);
}

/************************************************************************/
/* Procedure:  do_pause
 * Purpose:    to pause CD-ROM
 * 
 * Inputs:     program name, CD file des.
 * Outputs:    to CD-ROM, debug output
 * Returns:    0< on error; 0 (TRUE) on success
 * Notes:  
 *   1.
 */
/************************************************************************/
int do_pause(cdtool_t *cdt, int cdfd) 
{
    DEBUGMSG("do_pause: called, name=%s, cdfd=%d", cdt->progname, cdfd);
    VERBOSEMSG("Playing");

    if (ioctl(cdfd, CDROMPAUSE) == -1) {
        ERRORMSG("%s: ioctl cdrompause: %s", cdt->progname, strerror(errno));
        return(-1);
    }

    return(TRUE);
}

/************************************************************************/
/* Procedure:  do_stop
 * Purpose:    to stop playing a CD-ROM
 * 
 * Inputs:     program name, CD file des.
 * Outputs:    to CD-ROM, debug output
 * Returns:    0< on error; 0 (TRUE) on success
 * Notes:  
 *   1.
 */
/************************************************************************/
int do_stop(cdtool_t *cdt, int cdfd, int iCDNum) 
{
    char  caBuf[40];
    FILE  *fpPID;
    pid_t shuffle_pid;

    DEBUGMSG("do_stop: called, name=%s, cdfd=%d", cdt->progname, cdfd);
    VERBOSEMSG("Stopping");

    /* if shuffle is running, stop it... */
    snprintf(caBuf, sizeof(caBuf)-1, "%s%d", PID_FILE, iCDNum);
    if ((fpPID = fopen (caBuf, "r")) == NULL) {
        VERBOSEMSG("%s: Error opening pid file %s: %s", cdt->progname,
          caBuf, strerror(errno));
    } else {
	if (fgets (caBuf, sizeof(caBuf), fpPID) != NULL) {
	    shuffle_pid = atoi (caBuf);
	    kill (shuffle_pid, SIGINT);
	}
	fclose (fpPID);
    }

    /* send stop to the CDROM device */
    if (ioctl(cdfd, CDROMSTOP) == -1) {
	ERRORMSG("%s: ioctl cdromstop: %s", cdt->progname, strerror(errno));
	return(-1);
    }

    return(TRUE);
}

/************************************************************************/
/* Procedure:  do_eject
 * Purpose:    to eject the CD-ROM 
 * 
 * Inputs:     program name, CD file des.
 * Outputs:    to CD-ROM, debug output
 * Returns:    0< on error; 0 (TRUE) on success
 * Notes:  
 *   1.
 */
/************************************************************************/
int do_eject(cdtool_t *cdt, int cdfd)
{
    DEBUGMSG("do_eject:  called, name=%s, cdfd=%d", cdt->progname, cdfd);
    VERBOSEMSG("Ejecting CD-ROM");

    if (ioctl(cdfd, CDROMEJECT) == -1) {
        ERRORMSG("%s: ioctl cdromeject: %s", cdt->progname, strerror(errno));
        return(-1);
    }

    return(TRUE);
}

/************************************************************************/
/* Procedure:  do_close
 * Purpose:    to close the CD-ROM 
 * 
 * Inputs:     program name, CD file des.
 * Outputs:    to CD-ROM, debug output
 * Returns:    0< on error; 0 (TRUE) on success
 * Notes:  
 *   1.
 */
/************************************************************************/
int do_close(cdtool_t *cdt, int cdfd)
{
    DEBUGMSG("do_close: called, name=%s, cdfd=%d", cdt->progname, cdfd);
    VERBOSEMSG("Closing CD-ROM tray\n");

#ifndef __SOLARIS__
    if (ioctl(cdfd, CDROMCLOSETRAY) == -1) {
        ERRORMSG("%s: ioctl cdromclose: %s", cdt->progname, strerror(errno));
        return(-1);
    }
#endif

    return(TRUE);
}

/************************************************************************/
/* Procedure:  do_reset
 * Purpose:    to reset the CD-ROM
 * 
 * Inputs:     program name, CD file des.
 * Outputs:    to CD-ROM, debug output
 * Returns:    n/a
 * Notes:  
 *   1.
 */
/************************************************************************/
int do_reset(cdtool_t *cdt, int cdfd)
{
    DEBUGMSG("do_reset:  called, name=%s, cdfd=%d", cdt->progname, cdfd);
    VERBOSEMSG("Reseting CDROM\n");

#ifdef __SOLARIS__
    /* Solaris doesn't know CDROMRESET */
    if (ioctl(cdfd, CDROMSTOP) == -1) {
        ERRORMSG("%s: ioctl cdromstop: %s", cdt->progname, strerror(errno));
        return(-1);
    }
#else
    if (ioctl(cdfd, CDROMRESET) == -1) {
        ERRORMSG("%s: ioctl cdromreset: %s", cdt->progname, strerror(errno));
        return(-1);
    }
#endif

    if (ioctl(cdfd, CDROMSTART) == -1) {
        ERRORMSG("%s: ioctl cdromstart: %s", cdt->progname, strerror(errno));
        return(-1);
    }

    return(TRUE);
}

/************************************************************************/
/* Procedure:  do_volume
 * Purpose:    to adjust CD-ROM volume level
 * 
 * Inputs:     program name, CD file des., volume level (uchar)
 * Outputs:    to CD-ROM, to stdout
 * Returns:    0 on success, -1 on failure
 * Notes:  
 *   1.
 */
/************************************************************************/
int do_volume(cdtool_t *cdt, int cdfd, int level)
{
    struct cdrom_volctrl volctrl;

    DEBUGMSG("do_volume:  called, name=%s, volume=%d, cdfd=%d", cdt->progname,
      level, cdfd);

    if ((level < 0) || (level > 255)) {
        ERRORMSG("%s: illegal volume (allowed values 0-255)", cdt->progname);
        return -1;
    }

    VERBOSEMSG("Setting volume to level %d", level);

    volctrl.channel0=
        volctrl.channel1=
        volctrl.channel2=
        volctrl.channel3=level;

    if (ioctl(cdfd, CDROMVOLCTRL, &volctrl) == -1) {
        ERRORMSG("%s: CDROMVOLCTRL ioctl failed: %s\n", cdt->progname,
          strerror(errno));
        return -1;
    }

    return 0;
}

