/*
 * util.c
 *
 * Copyright 2004, Max Vozeler <max@hinterhof.net>
 * 
 */ 

#include "common.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include "util.h"

/* globals */
int cdtool_show_crlf = DOLF;
int cdtool_show_verbose = FALSE;
int cdtool_show_debug = FALSE;

/*
 *
 */

char *device_find(char *choice)
{
    char tmp[256];
    char *devname = NULL;
    struct device_name_t *m;
    int i;
    
    /*  device name, number of subdevices */
    static struct device_name_t {
        char *name;
        int subcount;
    } device_map[] = {
        { "/dev/cdrom", 1 },        /* Linux default */
        { "/dev/cdrom%d", 16 },     /* Other Linuxes? */
        { "/dev/scd%d", 16 },       /* Linux SCSI */
        { "/dev/sonycd", 1 },       /* Sony */
        { "/dev/mcd", 1 },          /* Mitsumi */
        { "/dev/sbpcd", 1 },        /* SoundBlaster Pro */ 
        { "/dev/sr%d", 16 },        /* Sun SCSI */
        { "/vol/dev/aliases/cdrom%d", 16 }, /* Sun Volume Manager */
        { NULL, 0 }
    };

    if (choice != NULL) {
        /* try user choice; display any error messages */
        devname = device_resolve(choice, 0);
        if (devname != NULL)
            return devname;
    }

    m = &device_map[0];
    while (m->name != NULL) {
        i = 0;
        while (m->subcount > i) {
            snprintf(tmp, sizeof(tmp)-1, m->name, i);
            tmp[sizeof(tmp)-1]=0;
            if ((devname = device_resolve(tmp, 1)) != NULL) {
                /* found */
                DEBUGMSG("using detected cd_device=%s", devname);
                return devname;
            }
            i++;
        }
        m++;
    }
 
    return NULL;
}

/*
 *
 */

char *device_resolve(char *path, int quiet)
{
    struct stat s;
    int level;
    char *linktarget;
    char *devname;
    char *devbase;
    char *splitter;
    size_t klen, nlen, ilen;

    level = 0;
    devname = strdup(path);

    while (++level < SYMLINK_MAX_DEPTH) {
        if (lstat(devname, &s) != 0) {
            /* not found */
            if (!quiet)
                ERRORMSG("%s not found.", devname);
            free(devname);
            return NULL;
        }

        /* something other than a link or block device? */
        if (!S_ISLNK(s.st_mode) && !S_ISBLK(s.st_mode)) {
            ERRORMSG("%s is not a block device.", devname);
            free(devname);
            return NULL;
        }

        if (S_ISLNK(s.st_mode)) {
            nlen = strlen(devname) + 1024 + 1;
            if ((devbase = malloc(nlen)) == NULL)
                return NULL;

            /* get full path from devname or getcwd() */
            if ((splitter = strrchr(devname, '/')) != NULL) {
                strncpy(devbase, devname, nlen-1);
                devbase[nlen-1] = 0;
                if ((splitter = strrchr(devbase, '/')) != NULL)
                    *++splitter = 0;
            } else {
                memset(devbase, 0, nlen);
                getcwd(devbase, nlen-1);
            }
     
            klen = strlen(devbase) + 1024 + 1;
            if ((linktarget = malloc(klen)) == NULL)
                return NULL;

            /* resolve the link */
            ilen = strlen(devbase);
            nlen = readlink(devname, &linktarget[ilen], klen - ilen - 1);
            if (nlen == -1) {
                ERRORMSG("%s not found: %s", devname, strerror(errno));
                free(devbase);
                free(devname);
                free(linktarget);
                return NULL;
            } else 
                linktarget[ilen + nlen] = 0;

            /* prepend base for relative links */
            if (linktarget[ilen] != '/') {
                /* doesnt write \0 */
                strncpy(&linktarget[0], devbase, strlen(devbase));
                ilen = 0;
            }
            
            /* next round with resolved name */
            free(devname);
            devname = strdup(&linktarget[ilen]);
            free(devbase);
            free(linktarget);
            devbase = NULL;
            linktarget = NULL;
            continue;
        }
        
        /* resolved to block device. name contains the full path */
        break;
    } 

    if (level >= SYMLINK_MAX_DEPTH) {
        ERRORMSG("too many recursions for %s", devname);
        return NULL;
    }

    DEBUGMSG("resolved device %s to %s", path, devname);
    return(devname);
}

/*
 * Solaris needs this
 *
 */

#ifndef HAVE_STRSEP
char *strsep(char **stringp, const char *delim)
{
    char *str = *stringp;
    char *end;

    if (!str)
        return NULL;

    end = strpbrk(str, delim);
    if (end)
        *end++ = '\0';

    *stringp = end;
    return str;
}
#endif

/*
 * 
 */

void do_crlf(FILE *term)
{
    static char *newline[] = { "", "\n", "\r\n" };
    fprintf(term, "%s", newline[cdtool_show_crlf]);
}


void bugfound(char *message, char *file, char *function, int line)
{
    struct utsname uts;
    uname(&uts);

    errormsg("\n"
     "BUG: %s at file %s, function %s (line %d)",
     message, file, function, line);

    errormsg(
     "BUG: The authors of this software think that you have just found\n"
     "     a bug in CDTOOL. We would appreciate you sending this notice\n"
     "     to cdtool-devel@lists.hinterhof.net, including if you can\n"
     "     some information (vendor, model) about the CDROM drive and\n"
     "     controller.");

    errormsg("\nCDTOOL version %s", PACKAGE_VERSION);
    errormsg("  running on %s %s %s (%s)\n",
     uts.machine,
     uts.sysname,
     uts.release,
     uts.version);
}

void usermsg(FILE *term, int crlf, char *fmt, va_list args)
{
    vfprintf(term, fmt, args);
    do_crlf(term);
}

void debugmsg(char *fmt, ...)
{
    va_list args;
    if (!cdtool_show_debug)
        return;
    va_start(args, fmt);
    usermsg(stderr, cdtool_show_crlf, fmt, args);
    va_end(args);
}

void errormsg(char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    usermsg(stderr, cdtool_show_crlf, fmt, args);
    va_end(args);
}

void verbosemsg(char *fmt, ...)
{
    va_list args;
    if (!cdtool_show_verbose)
        return;
    va_start(args, fmt);
    usermsg(stdout, cdtool_show_crlf, fmt, args);
    va_end(args);
}

void infomsg(char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    usermsg(stdout, cdtool_show_crlf, fmt, args);
    va_end(args);
}

